@@ -21,6 +21,9 @@ except Exception as e:
 
 FPrint = None
 
+# Exit with error on any exception, included those happening in async callbacks
+sys.excepthook = lambda *args: (traceback.print_exception(*args), sys.exit(1))
+
 def load_image(img):
     png = cairo.ImageSurface.create_from_png(img)
 
@@ -43,12 +46,15 @@ def load_image(img):
 
     return img
 
-if hasattr(os.environ, 'MESON_SOURCE_ROOT'):
-    root = os.environ['MESON_SOURCE_ROOT']
+if 'FP_PRINTS_PATH' in os.environ:
+    prints_path = os.environ['FP_PRINTS_PATH']
 else:
-    root = os.path.join(os.path.dirname(__file__), '..')
+    if 'MESON_SOURCE_ROOT' in os.environ:
+        root = os.environ['MESON_SOURCE_ROOT']
+    else:
+        root = os.path.join(os.path.dirname(__file__), '..')
 
-imgdir = os.path.join(root, 'examples', 'prints')
+    prints_path = os.path.join(root, 'examples', 'prints')
 
 ctx = GLib.main_context_default()
 
@@ -73,10 +79,12 @@ class VirtualImage(unittest.TestCase):
         assert cls.dev is not None, "You need to compile with virtual_image for testing"
 
         cls.prints = {}
-        for f in glob.glob(os.path.join(imgdir, '*.png')):
+        for f in glob.glob(os.path.join(prints_path, '*.png')):
             n = os.path.basename(f)[:-4]
             cls.prints[n] = load_image(f)
 
+        assert cls.prints, "No prints found in " + prints_path
+
     @classmethod
     def tearDownClass(cls):
         shutil.rmtree(cls.tmpdir)
@@ -234,7 +242,8 @@ class VirtualImage(unittest.TestCase):
         self.assertEqual(self.dev.get_finger_status(), FPrint.FingerStatusFlags.NONE)
         self.assertEqual(self._enrolled.props.driver, self.dev.get_driver())
         self.assertEqual(self._enrolled.props.device_id, self.dev.get_device_id())
-        self.assertEqual(self._enrolled.props.device_stored, self.dev.has_storage())
+        self.assertEqual(self._enrolled.props.device_stored,
+                         bool(self.dev.get_features() & FPrint.DeviceFeature.STORAGE))
         self.assertIsNone(self._enrolled.get_image())
 
         return self._enrolled
